﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace IngredConvert
{
    abstract internal class converterBase
    {
        protected double origQty = 0;
        protected int qtyWhole = 0;
        protected double remainder = 0;

        // gran: keep fractional parts as the main number, not as part of the remainder
        // eg: 8 means to keep 1/8s as a non-remainder. any remainder will be less than 1/8 (0.125)
        protected virtual int gran
        {
            get { return 1; }
        }

        // things happen in parent c'tors before children can init properties. use init() guarantee no conflict
        protected abstract void init();
        public converterBase(double qty)
        {
            init();

            origQty = qty * gran;
            qtyWhole = (int)origQty;
            remainder = (origQty - qtyWhole) / gran;

            // if the remainder is only a tiny fraction of the number of units, discard it
            if (origQty / 15 > remainder)
            {
                remainder = 0.0d;
                origQty = qtyWhole;
            }
        }

        protected static int div(ref int remaining, int m)
        {
            int quotient = remaining / m;
            remaining = remaining % m;
            return quotient;
        }

        public override string ToString()
        {
            string ret = toString().Trim();
            // since it's a double, test for tiny remainder instead of == 0
            if (!remaindersHandled() && remainder >= 0.001)
            {
                if (ret == "")
                    ret = "0";

                ret += String.Format(" (remainder {0:0.###} {1})", remainder, minWholeUnit());
            }
            return ret;
        }
        protected virtual bool remaindersHandled()
        {
            return false;
        }
        protected abstract string toString();
        protected abstract string minWholeUnit(); // ex: oz, tsp
    }

    abstract internal class cupsConverter : converterBase
    {
        public int tsp8 { get; protected set; } // 1/8 tsp
        public int tsp4 { get; protected set; }
        public int tsp2 { get; protected set; }
        public int tsp { get; protected set; }
        public int tbsp { get; protected set; }
        public int c { get; protected set; }

        public int tbsp2 { get; protected set; }

        protected override int gran
        {
            get { return 8; }
        }

        protected override void init()
        {
            // don't think this is needed for this class, but it's here to be safe
            tsp8 = tsp4 = tsp2 = tsp = tbsp = tbsp2 = c = 0;
        }
        public cupsConverter(double tsp)
            : base(tsp)
        {
            tsp8 = qtyWhole;
            convertCups();
            convertPartialCups();
            convertSpoons();
            convertCupsTbsp2();
        }
        void convertCups()
        {
            int remaining8 = tsp8;

            c = div(ref remaining8, 16 * 3 * 8); // 1/8 tsp in cup

            tsp8 = remaining8;
        }
        protected abstract void convertPartialCups();
        void convertSpoons()
        {
            int remaining8 = tsp8;

            tbsp = div(ref remaining8, 3 * 8); // 1/8 tsp in TBSP
            tsp = div(ref remaining8, 8); //    1 tsp
            tsp2 = div(ref remaining8, 4); // 1/2 tsp
            tsp4 = div(ref remaining8, 2); // 1/4 tsp

            tsp8 = remaining8;
        }
        void convertCupsTbsp2()
        {
            if (tsp == 1 && tsp2 == 1 && tsp4 == 0 && tsp8 == 0)
            {
                tsp = 0;
                tsp2 = 0;
                tbsp2 = 1;
            }
            else
                tbsp2 = 0;
        }

        protected abstract int getScoops();
        /// <summary>How many scoops are needed. For 2 cups + 1 tsp, Scoops is 3</summary>
        public int Scoops
        {
            get { return tsp8 + tsp4 + tsp2 + tsp + tbsp2 + tbsp + c + getScoops(); }
        }
        protected abstract int getCupSpoonCount();
        /// <summary>How many physical cups and spoons are used. For 2 cups + 1 tsp, CupSpoonCount is 2</summary>
        public int CupSpoonCount
        {
            get
            {
                return
                    (tsp8 > 0 ? 1 : 0) +
                    (tsp4 > 0 ? 1 : 0) +
                    (tsp2 > 0 ? 1 : 0) +
                    (tsp > 0 ? 1 : 0) +
                    (tbsp2 > 0 ? 1 : 0) +
                    (tbsp > 0 ? 1 : 0) +
                    (c > 0 ? 1 : 0) +
                    getCupSpoonCount();
            }
        }

        protected abstract string toStringPartialCups();
        protected override string toString()
        {
            string partial = toStringPartialCups();

            string cupsStr = (String.IsNullOrEmpty(partial))
                ? (c > 0 ? c + " c " : "")
                : (c > 0 ? c + "-" : "") + partial + "c ";

            string tbspStr = tbsp2 == 0
                ? (tbsp > 0 ? tbsp + " TBSP " : "")
                : (tbsp > 0 ? tbsp + "-" : "") + "1/2 TBSP ";

            int tsp8parts = tsp8 + tsp4 * 2 + tsp2 * 4;
            string tsp8Str = "";
            if (tsp8parts != 0)
            {
                int n = tsp8parts;
                int d = 8;
                while (n % 2 == 0)
                {
                    n /= 2;
                    d /= 2;
                }
                tsp8Str = n + "/" + d;
            }

            string tspStr = (String.IsNullOrEmpty(tsp8Str))
                ? (tsp > 0 ? tsp + " tsp" : "")
                : (tsp > 0 ? tsp + "-" : "") + tsp8Str + " tsp";


            return cupsStr + tbspStr + tspStr;
        }
        protected override string minWholeUnit()
        {
            return "tsp";
        }
    }

    // uses 1/4 cup increments
    internal class cups4Converter : cupsConverter
    {
        public int c4 { get; protected set; } //  1/4c
        public int c2 { get; protected set; } //  1/2c
        public int c34 { get; protected set; } // 3/4c

        protected override void init()
        {
            // if this is in c'tor, values set b4 c'tor will be overwritten with 0 in c'tor
            c4 = c2 = c34 = 0;
        }
        public cups4Converter(double tsp) : base(tsp) { }
        protected override void convertPartialCups()
        {
            int remaining8 = tsp8;

            c4 = div(ref remaining8, 4 * 3 * 8); // 1/8 tsp in 1/4 cup

            if (c4 == 2) // 1/2c
            {
                c4 = 0;
                c2 = 1;
            }
            if (c4 == 3) // 3/4c
            {
                c4 = 0;
                c34 = 1;
            }

            tsp8 = remaining8;
        }

        protected override int getScoops()
        {
            return c4 + c2 + c34;
        }

        protected override int getCupSpoonCount()
        {
            return c4 + c2 + c34;
        }

        protected override string toStringPartialCups()
        {
            if (c4 > 0)
                return "1/4";
            if (c2 > 0)
                return "1/2";
            if (c34 > 0)
                return "3/4";
            return "";
        }
    }

    // uses 1/3 cup increments
    internal class cups3Converter : cupsConverter
    {
        public int c3 { get; protected set; } //  1/3c
        public int c23 { get; protected set; } // 2/3c

        protected override void init()
        {
            // if this is in c'tor, values set b4 c'tor will be overwritten with 0 in c'tor
            c23 = 0;
        }
        public cups3Converter(double tsp) : base(tsp) { }
        protected override void convertPartialCups()
        {
            int remaining8 = tsp8;

            c3 = div(ref remaining8, 16 * 8); // 1/8 tsp in 1/3 cup

            if (c3 == 2) // 2/3c
            {
                c3 = 0;
                c23 = 1;
            }

            tsp8 = remaining8;
        }

        protected override int getScoops()
        {
            return c3 + c23;
        }

        protected override int getCupSpoonCount()
        {
            return c3 + c23;
        }

        protected override string toStringPartialCups()
        {
            if (c3 > 0)
                return "1/3";
            if (c23 > 0)
                return "2/3";
            return "";
        }
    }

    internal class poundsConverter : converterBase
    {
        public int oz8 { get; protected set; } // 1/8 oz
        public int oz4 { get; protected set; }
        public int oz38 { get; protected set; } // 3/8 oz
        public int oz2 { get; protected set; }
        public int oz58 { get; protected set; }
        public int oz34 { get; protected set; }
        public int oz78 { get; protected set; }
        public int oz { get; protected set; }
        public int lb { get; protected set; }

        // use only if there are no finer fractions
        public int lb4 { get; protected set; }
        public int lb2 { get; protected set; }
        public int lb34 { get; protected set; }

        protected override int gran
        {
            get { return 8; }
        }

        protected override void init()
        {
            // don't think this is needed for this class, but it's here to be safe
            oz8 = oz4 = oz38 = oz2 = oz58 = oz34 = oz78 = oz = lb = lb4 = lb2 = lb34 = 0;
        }
        public poundsConverter(double oz)
            : base(oz)
        {
            oz8 = qtyWhole;

            convertWeight();
            convertPartialLbs();
        }

        private void convertWeight()
        {
            int remaining8 = oz8;

            lb = div(ref remaining8, 16 * 8); // 1/8 oz in lb
            oz = div(ref remaining8, 8); //     1 oz
            oz78 = div(ref remaining8, 7); // 7/8 oz
            oz34 = div(ref remaining8, 6); // 3/4 oz
            oz58 = div(ref remaining8, 5); // 5/8 oz
            oz2 = div(ref remaining8, 4); //  1/2 oz
            oz38 = div(ref remaining8, 3); // 3/8 oz
            oz4 = div(ref remaining8, 2); //  1/4 oz

            oz8 = remaining8;
        }

        private void convertPartialLbs()
        {
            // 1/4, 1/2, 3/4 lb
            // use only if there are no remainder oz
            if (oz8 + oz4 + oz38 + oz2 + oz58 + oz34 + oz78 > 0)
                return;
            if (oz == 4)
            {
                oz = 0;
                lb4 = 1;
            }
            if (oz == 8)
            {
                oz = 0;
                lb2 = 1;
            }
            if (oz == 12)
            {
                oz = 0;
                lb34 = 1;
            }
        }

        protected override string toString()
        {
            string partial;

            // whole + fraction lb
            partial = "";
            if (lb4 > 0)
                partial = "1/4";
            if (lb2 > 0)
                partial = "1/2";
            if (lb34 > 0)
                partial = "3/4";

            string lbStr = (String.IsNullOrEmpty(partial))
                ? (lb > 0 ? lb + " lb " : "")
                : (lb > 0 ? lb + " " : "") + partial + " lb ";

            // whole + fraction oz
            partial = "";
            if (oz78 > 0)
                partial = "7/8";
            if (oz34 > 0)
                partial = "3/4";
            if (oz58 > 0)
                partial = "5/8";
            if (oz2 > 0)
                partial = "1/2";
            if (oz38 > 0)
                partial = "3/8";
            if (oz4 > 0)
                partial = "1/4";
            if (oz8 > 0)
                partial = "1/8";

            string ozStr = (String.IsNullOrEmpty(partial))
                ? (oz > 0 ? oz + " oz " : "")
                : (oz > 0 ? oz + " " : "") + partial + " oz";

            return lbStr + ozStr;
        }
        protected override string minWholeUnit()
        {
            return "oz";
        }
    }

    internal class miscConverter : converterBase
    {
        public string Units { get; set; }

        protected override void init() { }
        public miscConverter(double qty) : base(qty) { }

        protected override string toString()
        {
            // whole units
            if (remainder == 0)
                return qtyWhole + " " + Units;

            // standard fraction
            Fraction remainderFraction = Fraction.ToFraction(remainder);
            if (remainderFraction.Denominator <= 10 || remainderFraction.Denominator == 16)
                return ((qtyWhole > 0) ? qtyWhole + " " : "") + remainderFraction.ToString() + " " + Units;

            return String.Format("{0:0.###} {1}", origQty, Units);
        }
        protected override bool remaindersHandled()
        {
            return true;
        }

        protected override string minWholeUnit()
        {
            return Units;
        }
    }

    public class IngredientConverter
    {
        public enum Calc { Mult, Div }
        enum UnitType { None, Weight, Volume, Misc }
        #region regex
        // it's harder than it looks to combine <integer> and <fracint>
        const string PATTERN = @"
^(?<measurement>          # begin group
  [\s,]*                  # group separators and optional beginning space
  (
    (?<integer>\d+)\.?|   # integer
    (?<decimal>\d*\.\d+)| # decimal
    (
      (
        (?<fracint>\d+)   # integer preceding fraction
        [-\s\.]+          # ex: 1-1/2, 1 1/2 (decimal point for consistency with integer above)
      )?
      (?<numtor>\d+)      # numerator
      /
      (?<dentor>[1-9]\d*) # denominator. 0 not allowed
    )
  )
  \s*
  (?<unit>[a-zA-Z]+|\#)    # hash sign sometimes used for pounds
)+                         # whole group
";
        Regex regex = new Regex(PATTERN, RegexOptions.ExplicitCapture | RegexOptions.IgnorePatternWhitespace | RegexOptions.Compiled);
        #endregion

        List<string> cupsAbbrev = new List<string> { "c", "cup", "cups" };
        List<string> tbspAbbrev = new List<string> { "tb", "tbl", "tbsp", "tbsps", "table", "tables",
            "tablesp", "tablespoon", "tablespoons" };
        List<string> tspAbbrev = new List<string> { "tsp", "t", "tspoon", "tspoons", "teaspoon", "teaspoons" };
        // populated in c'tor from other lists
        List<string> volumeAbbrev = new List<string>();

        List<string> poundsAbbrev = new List<string> { "pound", "lb", "lbs", "#" };
        List<string> ozAbbrev = new List<string> { "ounce", "ounces", "oz" };
        // populated in c'tor from other lists
        List<string> weightAbbrev = new List<string>();

        string origValue; // orig user-entered value
        string origMatch; // what was matched of the user-entered value
        public string Match
        {
            get { return origMatch; }
        }

        UnitType unitType;
        string unit;

        // sum of qty.s in orig
        double oldQty;

        public IngredientConverter()
        {
            foreach (string s in cupsAbbrev) volumeAbbrev.Add(s);
            foreach (string s in tbspAbbrev) volumeAbbrev.Add(s);
            foreach (string s in tspAbbrev) volumeAbbrev.Add(s);

            foreach (string s in poundsAbbrev) weightAbbrev.Add(s);
            foreach (string s in ozAbbrev) weightAbbrev.Add(s);
        }
        public string Convert(testUnit t)
        {
            return Convert(t.input, t.calc, t.mod);
        }
        public string Convert(string input, Calc calc, double mod)
        {
            // reset
            unitType = UnitType.None;
            unit = "";
            oldQty = 0;

            origValue = input;

            if (!parse())
                return "";

            double newQty = (calc == Calc.Mult) ? oldQty * mod : oldQty / mod;

            // volume conversion
            if (unitType == UnitType.Volume)
            {
                cups3Converter c3 = new cups3Converter(newQty);
                cups4Converter c4 = new cups4Converter(newQty);

                if (c3.CupSpoonCount > c4.CupSpoonCount)
                    return c4.ToString();
                return c3.ToString();
            }

            // weight conversion
            if (unitType == UnitType.Weight)
                return new poundsConverter(newQty).ToString();

            // no unit found
            var miscConverter = new miscConverter(newQty);
            miscConverter.Units = unit;
            return miscConverter.ToString();
        }

        private bool parse()
        {
            // there can only be 1 match since the regex begins with ^
            Match match = regex.Match(origValue);
            if (!match.Success)
                return false;

            /*
            //test regex match
            List<string> l = new List<string>();
            for (int i = 0; i < match.Groups.Count; i++)
            {
                Group g = match.Groups[i];
                string name = regex.GroupNameFromNumber(i);
                //if (g.Captures.Count == 0)
                //    continue;
                l.Add(String.Format("[{0}({1})]={2}", name, i, g.ToString()));
                for (int j = 0; j < g.Captures.Count; j++)
                    l.Add(String.Format("  [{0}({1})]c[{2}]={3}", name, i, j, g.Captures[j].ToString()));
            }
            foreach (var ln in l) Console.WriteLine(ln);
            int setBreakpointHere = 1; return false;
             */

            // full match
            origMatch = match.Groups[0].ToString().Trim();

            #region explanation of: foreach (Capture c in match.Groups["measurement"].Captures)
            // must loop through each 'measurement' capture. else this can occur:
            //
            // 3 tbsp 1/2 tsp
            //
            //[measurement][0]=3 tbsp
            //[measurement][1]= 1/2 tsp   <-- index: 1
            //[integer][0]=3
            //[numtor][0]=1               <-- index: 0
            //[dentor][0]=2               <-- index: 0
            //[unit][0]=tbsp
            //[unit][1]=tsp               <-- index: 1
            #endregion
            foreach (Capture c in match.Groups["measurement"].Captures)
            {
                // fix
                match = regex.Match(c.ToString());
                int i = 0;
            //}

            //// orig:
            //int matches = match.Groups["measurement"].Captures.Count;
            //for (int i = 0; i < matches; i++)
            //{
                string integerStr = match.Groups["integer"].GroupCapture(i);
                string decimalStr = match.Groups["decimal"].GroupCapture(i);
                string fracintStr = match.Groups["fracint"].GroupCapture(i);
                string numtorStr = match.Groups["numtor"].GroupCapture(i);
                string dentorStr = match.Groups["dentor"].GroupCapture(i);
                // will always have a unit
                string unitStr = match.Groups["unit"].Captures[i].ToString();

                // get value
                double qty;
                if (!String.IsNullOrEmpty(integerStr))
                    qty = Int32.Parse(integerStr);
                else if (!String.IsNullOrEmpty(decimalStr))
                    qty = Double.Parse(decimalStr);
                else if (!String.IsNullOrEmpty(numtorStr))
                {
                    qty = Double.Parse(numtorStr) / Double.Parse(dentorStr);
                    if (!String.IsNullOrEmpty(fracintStr))
                        qty += Int32.Parse(fracintStr);
                }
                else
                    throw new Exception("Isn't it a comforting feeling to get the \"this error should never occur\" error? Sorry.");

                // string => UnitType
                UnitType tempUnitType;
                // capital T is tablespoons
                if (unitStr == "T")
                    unitStr = "tbsp";
                string unitLower = unitStr.ToLower();
                if (volumeAbbrev.Contains(unitLower))
                    tempUnitType = UnitType.Volume;
                else if (weightAbbrev.Contains(unitLower))
                    tempUnitType = UnitType.Weight;
                else
                    tempUnitType = UnitType.Misc;

                // if prev unitType isn't None, make sure they match. Misc must match exactly
                if (unitType != UnitType.None)
                    if (unitType != tempUnitType || (unitType == UnitType.Misc && unit != unitLower))
                        return false;
                unit = unitLower;
                unitType = tempUnitType;

                switch (unitType)
                {
                    case UnitType.Volume: // reduce to tsp
                        if (cupsAbbrev.Contains(unit))
                            qty *= 16 * 3;
                        else if (tbspAbbrev.Contains(unit))
                            qty *= 3;
                        // else teaspoons
                        else if (!tspAbbrev.Contains(unit))
                            throw new Exception("Volume unit not handled correctly by program. My fault, not yours. Sorry.");
                        break;
                    case UnitType.Weight: // reduce to oz
                        if (poundsAbbrev.Contains(unit))
                            qty *= 16;
                        // else oz
                        else if (!ozAbbrev.Contains(unit))
                            throw new Exception("Weight unit not handled correctly. My fault, not yours. Sorry.");
                        break;
                    default:
                        break;
                }

                oldQty += qty;
            }

            return true;
        }
    }

    internal static class RegexExt
    {
        public static string GroupCapture(this Group g, int captureIndex)
        {
            captureIndex = Math.Abs(captureIndex);
            // this will also handle 0
            if (g.Captures.Count <= captureIndex)
                return "";
            return g.Captures[captureIndex].ToString();
        }
    }

    public struct testUnit { public string input; public IngredConvert.IngredientConverter.Calc calc; public double mod; }
}
